#include <iostream>
#include <vector>
#include <cmath>
#include "hsImage.h"


#define IS_FOREGROUND 1.0f
#define IS_BACKGROUND 0.0f
#define IS_UNKNOWN 0.5f
#define X_AXIS 1
#define Y_AXIS 2
#define SEGMENT_SIZE_THRESHOLD 32
#define FIRST_PASS_ITERATION 10
#define SECOND_PASS_ITERATION 5
#define THIRD_PASS_ITERATION 20
#define LAMBDA 65536

//Pointer to global image
hsImage* GLOBAL_IMAGE;
int CURRENT_PROCESS_STATE;

struct kdNode{
	int x0, y0, x1, y1;
	int width, height;
	int Neighbor;
	int Parent;
	int LUChild, RDChild;
	int BaryX, BaryY;
	int DivisionAxis;
	hsImage* image;
	int numUnknown;
	bool hasF, hasB, hasU;

	kdNode(hsImage* _image, int _x0, int _y0, int _x1, int _y1, int _parent)
	{
		x0 = _x0;		x1 = _x1;
		y0 = _y0;		y1 = _y1;
		width = x1 - x0;
		height = y1 - y0;
		Parent = _parent;
		image = _image;

		LUChild = -1;
		RDChild = -1;
		Neighbor = -1;
		numUnknown = 0;
		DivisionAxis = -1;
		hasF = false; hasB = false; hasU = false;

		float bX = 0;
		float bY = 0;
		for(int y=y0; y<y1; y++){
			for(int x=x0; x<x1; x++)
			{
				if(image->constraint[y*image->width + x] == 1 && image->alphaMap[y*image->width + x] > 0.99f)
					hasF = true;
				if(image->constraint[y*image->width + x] == 1 && image->alphaMap[y*image->width + x] < 0.01f)
					hasB = true;
				else if(image->constraint[y*image->width + x] == 0)
				{
					hasU = true;
					numUnknown++;
					bX += x;
					bY += y;
				}
			}
		}

		BaryX = bX/numUnknown;
		BaryY = bY/numUnknown;
	}

	void computeVariance(int& xvariance, int& yvariance)
	{
		xvariance = 0;
		yvariance = 0;
		for(int y=y0; y<y1; y++)
			for(int x=x0; x<x1; x++)
				if(image->constraint[y*image->width + x] == 0)
				{
					xvariance += x;
					yvariance += y;
				}
		xvariance /= numUnknown;
		yvariance /= numUnknown;
	}

	bool isDivisible(int axis)
	{
		bool leftF, leftB, leftU;			leftF = leftB = leftU = false;
		bool rightF, rightB, rightU;		rightF = rightB = rightU = false;
		bool upF, upB, upU;					upF = upB = upU = false;
		bool downF, downB, downU;			downF = downB = downU = false;

		//Trivially rejected cases
		if(axis != X_AXIS && axis != Y_AXIS){
			printf("Something is wrong %d is supposed to be either 1 or 2\n",axis);
			exit(1);
		}

		if(width < SEGMENT_SIZE_THRESHOLD || height < SEGMENT_SIZE_THRESHOLD)
			return false;
		if( hasF && !hasB &&  hasU)
			return false;
		if(!hasF &&  hasB &&  hasU)
			return false;

		//Axis dependent rejection cases
		switch(axis)
		{
		case X_AXIS:
			for(int y=y0; y<y1; y++){
				for(int x=x0; x<BaryX; x++)
				{
					if(image->alphaMap[y*image->width + x] > 0.99f && image->constraint[y*image->width + x] == 1)
						leftF = true;
					if(image->alphaMap[y*image->width + x] < 0.01f && image->constraint[y*image->width + x] == 1)
						leftB = true;
					if(image->constraint[y*image->width + x] == 0)
						leftU = true;
				}
			}
			for(int y=y0; y<y1; y++){
				for(int x=BaryX; x<x1; x++)
				{
					if(image->alphaMap[y*image->width + x] > 0.99f && image->constraint[y*image->width + x] == 1)
						rightF = true;
					if(image->alphaMap[y*image->width + x] < 0.01f && image->constraint[y*image->width + x] == 1)
						rightB = true;
					if(image->constraint[y*image->width + x] == 0)
						rightU = true;
				}
			}
				if((leftF && !leftB && leftU) && (!rightF && rightB && rightU))
					return false;
				if((!leftF && leftB && leftU) && (rightF && !rightB && rightU))
					return false;
				break;
		case Y_AXIS:;
			for(int y=y0; y<BaryY; y++)
				for(int x=x0; x<x1; x++)
				{
					if(image->alphaMap[y*image->width + x] > 0.99f && image->constraint[y*image->width + x] == 1)
						upF = true;
					if(image->alphaMap[y*image->width + x] < 0.01f && image->constraint[y*image->width + x] == 1)
						upB = true;
					if(image->constraint[y*image->width + x] == 0)
						upU = true;
				}
				for(int y=BaryY; y<y1; y++)
					for(int x=x0; x<x1; x++)
					{
						if(image->alphaMap[y*image->width + x] > 0.99f && image->constraint[y*image->width + x] == 1)
							downF = true;
						if(image->alphaMap[y*image->width + x] < 0.01f && image->constraint[y*image->width + x] == 1)
							downB = true;
						if(image->constraint[y*image->width + x] == 0)
							downU = true;
					}
					if((upF && !upB && upU) && (!downF && downB && downU))
						return false;
					if((!upF && upB && upU) && (downF && !downB && downU))
						return false;
					break;
		}

		//All clear.
		return true;
	}

	void printStatus()
	{
		printf("x0: %d		y0: %d		x1: %d		y1: %d\n",x0,y0,x1,y1);
		printf("width: %d	height: %d	Parent: %d	Neighbor: %d\n", width, height, Parent, Neighbor);
		printf("BaryX: %d	BaryY: %d\n",BaryX,BaryY);
	}
};
std::vector<kdNode> kdTree;


//Functions
void loadInput(char* inputFile, char* trimapFile);
void divideImage();
//First pass
void FirstImageProcessing();

//Conjugate Gradient method related
void solveConjugateGradient(hsImage* sImage, bool hasInitial);
void multiplyDVector(int* v1, float* v2, float* result, int vSize);
void multiplyScalarVector(float scalar, float* v1, float* result, int vSize);
float transposeMultiplyVector(float* v0, float* v1, int vSize);
float getLength(float* v, int vSize);
void computeLp(hsImage* sImage, float* p, int radius, float* Lp);
void compute_ak(hsImage* sImage, float* p, int radius, int x, int y);

void visualizeGrayScaleVector(float* v, int width, int height, char* filename, bool DrawKDTree);

void loadInput(char* inputFile, char* trimapFile)
{
	//Load input image & trimap image
	IplImage* inputImage = cvLoadImage(inputFile);
	if(!inputImage)
	{
		std::cout << "Could not load input image: " << inputFile << std::endl;
		exit(1);
	}
	IplImage* trimapImage = cvLoadImage(trimapFile);
	if(!trimapImage)
	{
		std::cout << "Could not load trimap image: " << trimapFile << std::endl;
		exit(1);
	}

	if(!(inputImage->width == trimapImage->width && inputImage->height == trimapImage->height))
	{
		std::cout << "Input image resolution: " << inputImage->width << "X" << inputImage->height << std::endl;
		std::cout << "Trimap image resolution: " << trimapImage->width << "X" << trimapImage->height << std::endl;
		std::cout << "Input image resolution and trimap image resolution does not match!" << std::endl;
		exit(1);
	}

	GLOBAL_IMAGE = new hsImage((uchar*)inputImage->imageData,(uchar*)trimapImage->imageData,inputImage->width,inputImage->height,inputImage->nChannels);
}

void divideImage()
{
	int xvariance, yvariance;
	for(unsigned int i=0; i<kdTree.size(); i++)
	{
		kdTree[i].computeVariance(xvariance,yvariance);
		if(xvariance > yvariance && kdTree[i].isDivisible(X_AXIS))
		{
			kdTree[i].DivisionAxis = X_AXIS;
			kdTree[i].LUChild = kdTree.size();
			kdTree.push_back(kdNode(GLOBAL_IMAGE,kdTree[i].x0,kdTree[i].y0,kdTree[i].BaryX,kdTree[i].y1,i));
			kdTree[i].RDChild = kdTree.size();
			kdTree.push_back(kdNode(GLOBAL_IMAGE,kdTree[i].BaryX,kdTree[i].y0,kdTree[i].x1,kdTree[i].y1,i));
			//Set neighbor
			kdTree[kdTree.size()-2].Neighbor = kdTree.size()-1;
			kdTree[kdTree.size()-1].Neighbor = kdTree.size()-2;
		}
		else if(xvariance < yvariance && kdTree[i].isDivisible(Y_AXIS))
		{
			kdTree[i].DivisionAxis = Y_AXIS;
			kdTree[i].LUChild = kdTree.size();
			kdTree.push_back(kdNode(GLOBAL_IMAGE,kdTree[i].x0,kdTree[i].y0,kdTree[i].x1,kdTree[i].BaryY,i));
			kdTree[i].RDChild = kdTree.size();
			kdTree.push_back(kdNode(GLOBAL_IMAGE,kdTree[i].x0,kdTree[i].BaryY,kdTree[i].x1,kdTree[i].y1,i));
			//Set neighbor
			kdTree[kdTree.size()-2].Neighbor = kdTree.size()-1;
			kdTree[kdTree.size()-1].Neighbor = kdTree.size()-2;
		}
	}
}

void FirstImageProcessing()
{
	for(int i=0; i<kdTree.size(); i++)
	{
		if(kdTree[i].LUChild == -1)
		{
			hsImage* subImage = new hsImage(GLOBAL_IMAGE,kdTree[i].x0,kdTree[i].y0,kdTree[i].x1,kdTree[i].y1);
			solveConjugateGradient(subImage,false);
			int s=0;
			for(int y=kdTree[i].y0; y<kdTree[i].y1; y++){
				for(int x=kdTree[i].x0; x<kdTree[i].x1; x++){
					GLOBAL_IMAGE->vX[y*GLOBAL_IMAGE->width + x] = subImage->vX[s];
					GLOBAL_IMAGE->vP[y*GLOBAL_IMAGE->width + x] = subImage->vP[s];
					GLOBAL_IMAGE->vR[y*GLOBAL_IMAGE->width + x] = subImage->vR[s];
					s++;
				}
			}
		}
	}
}

void SecondImageProcessing()
{
	solveConjugateGradient(GLOBAL_IMAGE,false);
}

void ThirdImageProcessing()
{
	for(int i=0; i<kdTree.size(); i++)
	{
		if(kdTree[i].LUChild == -1)
		{
			hsImage* subImage = new hsImage(GLOBAL_IMAGE,kdTree[i].x0,kdTree[i].y0,kdTree[i].x1,kdTree[i].y1);
			solveConjugateGradient(subImage,true);
			int s=0;
			for(int y=kdTree[i].y0; y<kdTree[i].y1; y++){
				for(int x=kdTree[i].x0; x<kdTree[i].x1; x++){
					GLOBAL_IMAGE->vX[y*GLOBAL_IMAGE->width + x] = subImage->vX[s];
					GLOBAL_IMAGE->vP[y*GLOBAL_IMAGE->width + x] = subImage->vP[s];
					GLOBAL_IMAGE->vR[y*GLOBAL_IMAGE->width + x] = subImage->vR[s];
					s++;
				}
			}
		}
	}
}


void solveConjugateGradient(hsImage* sImage, bool hasInitial)
{
	int imageSize = sImage->numPixels;
	float* b = new float[imageSize];
	float* Lp = new float[imageSize];
	float* Ap = new float[imageSize];
	float* Dp = new float[imageSize];
	float AlphaK, BetaK;
	float rsold, rsnew;
	int NUM_ITERATION;

	//From sImage
	float* x = sImage->vX;
	float* p = sImage->vP;
	float* r = sImage->vR;
	
	//window size related
	int wr; //window radius
	int wb; //bandwidth
	switch(CURRENT_PROCESS_STATE)
	{
	case 1:
		wb = sImage->numPixels / std::max((sImage->uX1-sImage->uX0),(sImage->uY1-sImage->uY0));
		wr = 5;
		NUM_ITERATION = 10;
		break;
	case 2:
		//wr = 5;
		//wr = std::min(sImage->width,sImage->height)/50;
		wr = 20;
		NUM_ITERATION = 100;
		break;
	case 3:
		wb = sImage->numPixels / std::max((sImage->uX1-sImage->uX0),(sImage->uY1-sImage->uY0));
		wr = wb / 15;
		NUM_ITERATION = 20;
		break;
	default:
		printf("Invalid process state: %d\n",CURRENT_PROCESS_STATE);
		exit(0);
		break;
	}
	if(wr < 1)
		wr = 1;


	//Prepare for Lp calculation
	//printf("calculate mean pixel\n");
	if(sImage->meanPixel == NULL)
		sImage->calculateMeanPixel(wr);
	//printf("calculate covariance matrix\n");
	if(sImage->covarianceMatrix == NULL)
		sImage->calculateCovarianceMatrices(wr);


	//Compute b
	for(int i=0; i<imageSize; i++)
		b[i] = LAMBDA * sImage->constraint[i] * sImage->alphaMap[i];

	//Compute initial r and p
	if(!hasInitial)
	{
		//Use b-Ax0 = b- 0 = b as initial r and p
		for(int i=0; i<imageSize; i++)
		{
			x[i] = 0;
			r[i] = b[i];
			p[i] = b[i];
		}
	}
	/*
	else
	{
		//Compute Ax
		multiplyDVector(sImage->constraint,x,Dp,imageSize);
		multiplyScalarVector(LAMBDA,Dp,Dp,imageSize);
		sImage->calculateAkstarBkstar(wr,x);
		computeLp(sImage,x,wr,Lp);
		for(int i=0; i<imageSize; i++)
			Ap[i] = Lp[i] + Dp[i];
		//get b and r
		for(int i=0; i<imageSize; i++)
		{
			r[i] = b[i] - Ap[i];
			p[i] = r[i];
		}
	}
	*/
	
	//Main loop
	printf("\n");
	rsold = transposeMultiplyVector(r,r,imageSize);
	char buffer[100];
	for(int i=0; i<NUM_ITERATION; i++)
	{
		printf("%d ",i);
		//Update Ak-star and Bk-star according to change in p
		sImage->calculateAkstarBkstar(wr,p);

		//Compute Ap
		multiplyDVector(sImage->constraint,p,Dp,imageSize);
		multiplyScalarVector(LAMBDA,Dp,Dp,imageSize);
		computeLp(sImage,p,wr,Lp);
		for(int i=0; i<imageSize; i++)
			Ap[i] = Lp[i] + Dp[i];
		//Compute AlphaK
		AlphaK = rsold / transposeMultiplyVector(p,Ap,imageSize);
		//Compute x_k+1
		for(int i=0; i<imageSize; i++)
			x[i] = x[i] + AlphaK * p[i];
		//Compute r_k+1
		for(int i=0; i<imageSize; i++)
			r[i] = r[i] - AlphaK * Ap[i];
		//Compute rsnew
		rsnew = transposeMultiplyVector(r,r,imageSize);
		////////////////////////////////////////////////////////////////////
		//Skip checking r's size, since we have fixed number of iteration//
		//////////////////////////////////////////////////////////////////

		//Compute BetaK
		BetaK = rsnew/rsold;
		//Compute p_k+1
		for(int i=0; i<imageSize; i++)
			p[i] = r[i] + BetaK*p[i];

		//Swap rsold and rsnew at the end of loop
		rsold = rsnew;
		sprintf(buffer, "%d.bmp", i);
		visualizeGrayScaleVector(x, sImage->width, sImage->height, buffer, false);
	}
	printf("\n");

	//free all alocated resources
	delete[] b;
	delete[] Lp;
	delete[] Ap;
	delete[] Dp;
}

void multiplyDVector(int* v1, float* v2, float* result, int vSize)
{
	for(int i=0; i<vSize; i++)
		result[i] = v1[i] * v2[i];
}

void multiplyScalarVector(float scalar, float* v1, float* result, int vSize)
{
	for(int i=0; i<vSize; i++)
		result[i] = scalar * v1[i];
}

float getLength(float* v1, int vSize)
{
	float result = 0.0f;
	for(int i=0; i<vSize; i++)
		result += v1[i]*v1[i];
	return sqrt(result);
}

float transposeMultiplyVector(float* v0, float* v1, int vSize)
{
	float result = 0.0f;
	for(int i=0; i<vSize; i++)
		result += v0[i] * v1[i];
	return result;
}

void computeLp(hsImage* sImage, float* p, int radius, float* Lp)
{
	//Main loop for Lp
#pragma omp parallel for
	for(int _i=0; _i<sImage->numPixels; _i++)
	{
		int x = _i % sImage->width;
		int y = _i / sImage->width;

		int x0 = (x-radius >= 0) ? x-radius : 0;
		int y0 = (y-radius >= 0) ? y-radius : 0;
		int x1 = (x+radius <= sImage->width) ? x+radius : sImage->width;
		int y1 = (y+radius <= sImage->height) ? y+radius : sImage->height;
		int _w_ = (x1-x0) * (y1-y0);

		//First term
		Lp[_i] = _w_ * p[_i];
		//Get second term
		float second_term = 0;
		float sigmaAk[3] = {0.0f,0.0f,0.0f};
		float sigmaBk = 0.0f;
		for(int j=y0; j<y1; j++){
			for(int i=x0; i<x1; i++){
				sigmaBk += sImage->BkStar[j*sImage->width + i];
				sigmaAk[0] += sImage->AkStar[j*sImage->width + i][0];
				sigmaAk[1] += sImage->AkStar[j*sImage->width + i][1];
				sigmaAk[2] += sImage->AkStar[j*sImage->width + i][2];
			}
		}
		second_term = sigmaAk[0]*sImage->iPixel[_i*sImage->nChannels + 0] + sigmaAk[1]*sImage->iPixel[_i*sImage->nChannels + 1] + sigmaAk[2]*sImage->iPixel[_i*sImage->nChannels + 2];
		second_term += sigmaBk;
		
		//Combine
		Lp[_i] = Lp[_i] - second_term;
	}
}

void visualizeGrayScaleVector(float* v, int width, int height, char* filename, bool DrawKDTree)
{
	IplImage* fImg = cvCreateImage(cvSize(width, height),IPL_DEPTH_8U,1);
	for(int i=0; i<width*height; i++)
		fImg->imageData[i] = v[i]*255.0f;

	if(DrawKDTree == true){
		for(int i=0; i<kdTree.size(); i++)
			if(kdTree[i].LUChild == -1)
			{
				for(int x=kdTree[i].x0; x<kdTree[i].x1; x++)
				{
					fImg->imageData[kdTree[i].y0 * width + x] = 70;
					fImg->imageData[kdTree[i].y1 * width + x] = 70;
				}
				for(int y=kdTree[i].y0; y<kdTree[i].y1; y++)
				{
					fImg->imageData[y * width + kdTree[i].x0] = 70;
					fImg->imageData[y * width + kdTree[i].x1] = 70;
				}
			}
	}

	cvSaveImage(filename,fImg);
	cvReleaseImage(&fImg);
}

int main(int argc, char** argv)
{
	char aFilename[] = "alphamap.bmp";
	if(argc != 3)
	{
		std::cout << "Invalid argument! We take two argument <Input image filename> <Trimap filename>" << std::endl;
		exit(1);
	}

	//Load input image & input trimap
	loadInput(argv[1],argv[2]);

	//Create kd-tree
	kdTree.clear();
	kdTree.reserve(1024);
	kdTree.push_back(kdNode(GLOBAL_IMAGE,0,0,GLOBAL_IMAGE->width,GLOBAL_IMAGE->height,-1));
	divideImage();

	//visualizeGrayScaleVector(GLOBAL_IMAGE->alphaMap,GLOBAL_IMAGE->width,GLOBAL_IMAGE->height,"temporary.bmp");

	//1st Local pass
	/*
	CURRENT_PROCESS_STATE = 1;
	FirstImageProcessing();
	visualizeGrayScaleVector(GLOBAL_IMAGE->vX,GLOBAL_IMAGE->width,GLOBAL_IMAGE->height,"firstPass.bmp",false);
	*/
	//1st Global pass
	CURRENT_PROCESS_STATE = 2;
	SecondImageProcessing();
	visualizeGrayScaleVector(GLOBAL_IMAGE->vX,GLOBAL_IMAGE->width,GLOBAL_IMAGE->height,"secondPass.bmp",false);
	/*
	//2nd Local pass
	CURRENT_PROCESS_STATE = 3;
	ThirdImageProcessing();
	visualizeGrayScaleVector(GLOBAL_IMAGE->vX,GLOBAL_IMAGE->width,GLOBAL_IMAGE->height,"thirdPass.bmp",false);
	*/
}